"""
widgetselecter Module

this contains the class which allows the ViewConfig to select the appropriate widget for the given field

Classes:
Name                               Description
WidgetSelecter                     Parent Class
SAWidgetSelector                   Selecter Based on sqlalchemy field types
DatabaseViewWidgetSelector         Database View always selects the same widget
TableDefWidgetSelector             Table def fields use the same widget

Exceptions:
None

Functions:
None


Copyright (c) 2007-10 Christopher Perkins
Original Version by Christopher Perkins 2007Database
Released under MIT license.
"""
import warnings

from sprox._widgetselector import *

try:
    from sprox.sa.widgetselector import SAWidgetSelector as _SAWidgetSelector

    class SAWidgetSelector(_SAWidgetSelector):
        def __init__(self, *args, **kw):
            warnings.warn('This class has moved to the sprox.sa.widgetselector module.') # pragma: no cover
            _SAWidgetSelector.__init__(self, *args, **kw) # pragma: no cover
except ImportError: # pragma: no cover
    pass # pragma: no cover





"""
mingprovider Module

This contains the class which allows sprox to interface with any database.

Copyright &copy 2009 Jorge Vargas
Original Version by Jorge Vargas 2009
Released under MIT license.
"""
from bson.errors import InvalidId
import itertools
from sprox.iprovider import IProvider
from sprox.util import timestamp
import datetime, inspect

try:
    from ming.odm import mapper, ForeignIdProperty, FieldProperty, RelationProperty
    from ming.odm.declarative import MappedClass
    from ming.odm.property import OneToManyJoin, ManyToOneJoin, ORMProperty
    from ming.odm.icollection import InstrumentedObj
except ImportError: #pragma nocover
    from ming.orm import mapper, ForeignIdProperty, FieldProperty, RelationProperty
    from ming.orm.declarative import MappedClass
    from ming.orm.property import OneToManyJoin, ManyToOneJoin, ORMProperty
    from ming.orm.icollection import InstrumentedObj


from ming import schema as S
import bson
from bson import ObjectId

import re
from .widgetselector import MingWidgetSelector
from .validatorselector import MingValidatorSelector
from pymongo import ASCENDING, DESCENDING
from sprox._compat import string_type, zip_longest

class MingProvider(IProvider):
    default_view_names = ['_name', 'name', 'description', 'title']
    default_widget_selector_type = MingWidgetSelector
    default_validator_selector_type = MingValidatorSelector

    def __init__(self, hint, **hints):
        self.session = hint

    def get_field(self, entity, name):
        """Get a field with the given field name."""
        return mapper(entity).property_index[name]

    def get_fields(self, entity):
        """Get all of the fields for a given entity."""
        if inspect.isfunction(entity):
            entity = entity()
        return [prop.name for prop in mapper(entity).properties if isinstance(prop, ORMProperty)]

    @property
    def _entities(self):
        entities = getattr(self, '__entities', None)
        if entities is None:
            entities = dict(((m.mapped_class.__name__, m) for m in MappedClass._registry.values()))
            self.__entities = entities
        return entities
    
    def get_entity(self, name):
        """Get an entity with the given name."""
        return self._entities[name].mapped_class

    def get_entities(self):
        """Get all entities available for this provider."""
        return iter(self._entities.keys())

    def get_primary_fields(self, entity):
        """Get the fields in the entity which uniquely identifies a record."""
        return [self.get_primary_field(entity)]

    def get_primary_field(self, entity):
        """Get the single primary field for an entity"""
        return '_id'

    def _get_meta(self, entity, field_name, metaprop):
        """Returns the value of the given sprox meta property for the field."""
        field = self.get_field(entity, field_name)
        return getattr(field, "sprox_meta", {}).get(metaprop, None)

    def get_view_field_name(self, entity, possible_names, item=None):
        """Get the name of the field which first matches the possible colums

        :Arguments:
          entity
            the entity where the field is located
          possible_names
            a list of names which define what the view field may contain.  This allows the first
            field that has a name in the list of names will be returned as the view field.
        """
        if entity is InstrumentedObj:
            # Cope with subdocuments
            if item is not None:
                fields = list(item.keys())
            else:
                fields = ['_impl']
        else:
            fields = self.get_fields(entity)
            for field in fields:
                if self._get_meta(entity, field, 'title'):
                    return field

        view_field = None
        for column_name in possible_names:
            for actual_name in fields:
                if column_name == actual_name:
                    view_field = actual_name
                    break
            if view_field:
                break
            for actual_name in fields:
                if column_name in actual_name:
                    view_field = actual_name
                    break
            if view_field:
                break

        if view_field is None:
            view_field = fields[0]

        return view_field

    def get_dropdown_options(self, entity_or_field, field_name, view_names=None):
        """Get all dropdown options for a given entity field.

        :Arguments:
          entity_or_field
            either the entity where the field is located, or the field itself
          field_name
            if the entity is specified, name of the field in the entity. Otherwise, None
          view_names
            a list of names which define what the view field may contain.  This allows the first
            field that has a name in the list of names will be returned as the view field.

        :Returns:
          A list of tuples with (id, view_value) as items.

        """
        if view_names is None:
            view_names = self.default_view_names

        if field_name is not None:
            field = self.get_field(entity_or_field, field_name)
        else:
            field = entity_or_field

        if isinstance(field, FieldProperty):
            field_type = getattr(field, 'field_type', None)
            if field_type is None:
                f = getattr(field, 'field', None)
                if f is not None:
                    field = field.field
                    field_type = field.type

            schemaitem = field_type
            if isinstance(schemaitem, S.OneOf):
                return [ (opt,opt) for opt in schemaitem.options ]
            raise NotImplementedError("get_dropdown_options doesn't know how to get the options for field %r of type %r" % (field, schemaitem))

        if not isinstance(field, RelationProperty):
            raise NotImplementedError("get_dropdown_options expected a FieldProperty or RelationProperty field, but got %r" % field)
        try:
            join = field.join
            iter = join.rel_cls.query.find()
            rel_cls = join.rel_cls
        #this seems like a work around for a bug in ming.
        except KeyError: # pragma: no cover
            join = field.related
            iter = join.query.find()
            rel_cls = join

        view_field = self.get_view_field_name(rel_cls, view_names)
        return [ (str(obj._id), getattr(obj, view_field)) for obj in iter ]

    def get_relations(self, entity):
        """Get all of the field names in an enity which are related to other entities."""
        return [prop.name for prop in mapper(entity).properties if isinstance(prop, RelationProperty)]

    def is_relation(self, entity, field_name):
        """Determine if a field is related to a field in another entity."""
        return isinstance(self.get_field(entity, field_name), RelationProperty)

    def is_query(self, entity, value):
        """determines if a field is a query instead of actual list of data"""
        #Currently not supported in MING
        return False

    def is_nullable(self, entity, field_name):
        """Determine if a field is nullable."""
        fld = self.get_field(entity, field_name)
        if isinstance(fld, RelationProperty):
            # check the required attribute on the corresponding foreign key field
            fld = fld.join.prop
        return not getattr(fld, 'kwargs', {}).get("required", False)

    def get_field_default(self, field):
        field = getattr(field, 'field', None)
        if field is not None:
            if_missing = field.schema.if_missing
            if if_missing is not None:
                return (True, if_missing)
        return (False, None)

    def get_field_provider_specific_widget_args(self, entity, field, field_name):
        return {}

    def get_default_values(self, entity, params):
        return params

    def _related_object_id(self, value):
        if isinstance(value, MappedClass):
            return value._id
        return ObjectId(value)

    def _cast_value(self, entity, key, value):
        # handles the case where an record with no id is being created
        if key == '_id' and value == '':
            value = ObjectId()

        if value is None:
            # Let none pass as is as it actually means a "null" on mongodb
            return value
            
        field = getattr(entity, key)
        
        relations = self.get_relations(entity)
        if key in relations:
            related = field.related
            if isinstance(value, list):
                return related.query.find({'_id':{'$in':[self._related_object_id(i) for i in value]}}).all()
            else:
                return self.get_obj(related, {'_id':self._related_object_id(value)})

        field = getattr(field, 'field', None)
        if field is not None:
            if field.type is S.DateTime or field.type is datetime.datetime:
                if isinstance(value, string_type):
                    return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
                else:
                    return value
            elif field.type is S.Binary:
                return bson.Binary(value)
            elif field.type in (S.Int, int):
                return int(value)
            elif field.type in (S.Bool, bool):
                if value in ('true', 'false'):
                    return value == 'true' and True or False
                else:
                    return bool(value)
        return value

    def create(self, entity, params):
        """Create an entry of type entity with the given params."""
        values = {}
        fields = self.get_fields(entity)
        for key, value in params.items():
            if key not in fields:
                continue
            value = self._cast_value(entity, key, value)
            if value is not None:
                values[key] = value

        obj = entity(**values)
        self.flush()
        return obj

    def flush(self):
        self.session.flush_all()
        self.session.close_all()

    def get_obj(self, entity, params, fields=None, omit_fields=None):
        try:
            return entity.query.get(_id=ObjectId(params['_id']))
        except InvalidId:
            return None
        except KeyError:
            return entity.query.find_by(**params).first()

    def get(self, entity, params, fields=None, omit_fields=None):
        return self.dictify(self.get_obj(entity, params), fields, omit_fields)

    def update(self, entity, params, omit_fields=None):
        """Update an entry of type entity which matches the params."""
        obj = self.get_obj(entity, params)
        params.pop('_id')
        try:
            params.pop('sprox_id')
        except KeyError:
            pass
        try:
            params.pop('_method')
        except KeyError:
            pass

        fields = self.get_fields(entity)
        for key, value in params.items():
            if key not in fields:
                continue

            if omit_fields and key in omit_fields:
                continue

            value = self._cast_value(entity, key, value)
            setattr(obj, key, value)

        self.flush()
        return obj

    def delete(self, entity, params):
        """Delete an entry of typeentity which matches the params."""
        obj = self.get_obj(entity, params)
        obj.delete()
        return obj

    def _modify_params_for_related_searches(self, entity, params, view_names=None, substrings=()):
        if view_names is None:
            view_names = self.default_view_names

        relations = self.get_relations(entity)
        for relation in relations:
            if relation in params:
                value = params[relation]
                if not isinstance(value, string_type):
                    # When not a string consider it the related class primary key
                    params.update({relation: value})
                    continue

                if not value:
                    # As we use ``contains``, searching for an empty text
                    # will just lead to all results so we just remove the filter.
                    del params[relation]
                    continue

                relationship = getattr(entity, relation)
                target_class = relationship.related
                view_name = self.get_view_field_name(target_class, view_names)

                if relation in substrings:
                    filter = {view_name: {'$regex': re.compile(re.escape(value), re.IGNORECASE)}}
                else:
                    filter = {view_name: value}
                value = target_class.query.find(filter).all()
                params[relation] = value

        return params

    def _modify_params_for_relationships(self, entity, params):
        relations = self.get_relations(entity)
        for relation in relations:
            if relation in params:
                relationship = getattr(entity, relation)
                value = params[relation]

                if not isinstance(value, list):
                    value = [value]

                adapted_value = []
                for v in value:
                    if isinstance(v, ObjectId) or isinstance(v, string_type):
                        obj = self.get_obj(relationship.related, dict(_id=v))
                        if obj is not None:
                            adapted_value.append(obj)
                    else:
                        adapted_value.append(v)
                value = adapted_value

                join = relationship.join
                my_foreign_key = relationship._detect_foreign_keys(relationship.mapper,
                                                                   join.rel_cls,
                                                                   False)
                rel_foreign_key = relationship._detect_foreign_keys(mapper(relationship.related),
                                                                    join.own_cls,
                                                                    False)

                params.pop(relation)
                if my_foreign_key:
                    my_foreign_key = my_foreign_key[0]
                    params[my_foreign_key.name] = {'$in': [r._id for r in value]}
                elif rel_foreign_key:
                    rel_foreign_key = rel_foreign_key[0]
                    value = [getattr(r, rel_foreign_key.name) for r in value]
                    if rel_foreign_key.uselist:
                        value = list(itertools.chain(*value))
                    params['_id'] = {'$in': value}
        return params

    def query(self, entity, limit=None, offset=0, limit_fields=None,
              order_by=None, desc=False, filters={},
              substring_filters=[], search_related=False, related_field_names=None,
              **kw):

        if '_id' in filters:
            try:
                filters['_id'] = ObjectId(filters['_id'])
            except InvalidId:
                pass

        if search_related:
            # Values for related fields contain the text to search
            filters = self._modify_params_for_related_searches(entity, filters,
                                                               view_names=related_field_names,
                                                               substrings=substring_filters)
        filters = self._modify_params_for_relationships(entity, filters)

        for field in substring_filters:
            if self.is_string(entity, field):
                filters[field] = {'$regex': re.compile(re.escape(filters[field]), re.IGNORECASE)}

        iter = entity.query.find(filters)
        if offset:
            iter = iter.skip(int(offset))
        if limit is not None:
            iter = iter.limit(int(limit))

        if order_by is not None:
            if not isinstance(order_by, (tuple, list)):
                order_by = [order_by]

            if not isinstance(desc, (tuple, list)):
                desc = [desc]

            sorting = [(field, DESCENDING if sort_descending else ASCENDING) for field, sort_descending in
                       zip_longest(order_by, desc)]
            iter.sort(sorting)

        count = iter.count()
        return count, iter.all()

    def is_string(self, entity, field_name):
        fld = self.get_field(entity, field_name)
        if isinstance(fld, RelationProperty):
            # check the required attribute on the corresponding foreign key field
            fld = fld.join.prop

        fld = getattr(fld, 'field', None)
        return isinstance(fld.schema, S.String)

    def is_binary(self, entity, field_name):
        fld = self.get_field(entity, field_name)
        if isinstance(fld, RelationProperty):
            # check the required attribute on the corresponding foreign key field
            fld = fld.join.prop

        fld = getattr(fld, 'field', None)
        return isinstance(fld.schema,S.Binary)

    def relation_fields(self, entity, field_name):
        field = self.get_field(entity, field_name)
        if not isinstance(field, RelationProperty):
            raise TypeError("The field %r is not a relation field" % field)

        #This is here for many-to-many turbogears-ming relations
        if not field.join.prop:
            return []

        return [field.join.prop.name]

    def relation_entity(self, entity, field_name):
        """If the field in the entity is a relation field, then returns the
        entity which it relates to.

        :Returns:
          Related entity for the field
        """
        field = self.get_field(entity, field_name)
        return field.related

    def get_field_widget_args(self, entity, field_name, field):
        args = {}
        args['provider'] = self
        args['nullable'] = self.is_nullable(entity, field_name)
        return args

    def is_unique(self, entity, field_name, value):
        iter = entity.query.find({ field_name: value })
        return iter.count() == 0

    def is_unique_field(self, entity, field_name):
        for idx in getattr(entity.__mongometa__, "unique_indexes", ()):
            if idx == (field_name,):
                return True
        return False

    def dictify(self, obj, fields=None, omit_fields=None):
        if obj is None:
            return {}
        r = {}
        for prop in self.get_fields(obj.__class__):
            if fields and prop not in fields:
                continue

            if omit_fields and prop in omit_fields:
                continue

            value = getattr(obj, prop)
            if value is not None:
                if self.is_relation(obj.__class__, prop):
                    klass = self.relation_entity(obj.__class__, prop)
                    pk_name = self.get_primary_field(klass)
                    if isinstance(value, list):
                        #joins
                        value = [getattr(value, pk_name) for value in value]
                    else:
                        #fks
                        value = getattr(value, pk_name)
            r[prop] = value
        return r



"""
validatorselecter Module

this contains the class which allows the ViewConfig to select the appropriate validator for the given field

Classes:
Name                               Description
MongoValidatorSelecter             Class to aid in the selection of a validator for a mongo document field

Exceptions:
None

Functions:
None


Copyright (c) 2009 Christopher Perkins
Original Version by Christopher Perkins 2009
Released under MIT license.

"""

from sprox.validatorselector import ValidatorSelector
from ming import schema as s
try:
    import ming.odm as o
except ImportError: #pragma: no cover
    import ming.orm as o
import inspect

from formencode.validators import StringBool, Number, UnicodeString as FEUnicodeString, Email, Int
try: #pragma: no cover
    import tw2.forms
    from tw2.core.validation import *
    class UnicodeString(FEUnicodeString):
        outputEncoding = None
except ImportError: #pragma: no cover
    from tw.forms.validators import *
    DateTimeValidator = DateValidator
    BoolValidator = StringBool

class MingValidatorSelector(ValidatorSelector):

    default_validators = {
    s.Bool: BoolValidator,
    s.Int: Int,
    s.Float: Number,
    s.DateTime: DateTimeValidator,
    s.Binary: None,
    s.Value: None,
    s.ObjectId: None,
    s.String: UnicodeString
    }

    def select(self, field):

        #xxx: make this a "one of" validator
        if isinstance(field, o.RelationProperty):
            return UnicodeString

        schema_type = getattr(field, 'schema', None)
        field_type = getattr(field, 'field_type', None)
        if field_type is None:
            f = getattr(field, 'field', None)
            if f is not None:
                field = field.field
                field_type = field.type
                schema_type = field.schema

        type_ = s.String
        for t in list(self.default_validators.keys()):
            if isinstance(field_type, s.OneOf):
                break
            if inspect.isclass(field_type) and issubclass(field_type, t):
                type_ = t
                break
            if schema_type is not None and isinstance(schema_type, t):
                type_ = t
                break

        validator_type = self.default_validators.get(type_, None)
        return validator_type



"""
validatorselecter Module

this contains the class which allows the ViewConfig to select the appropriate validator for the given field

Classes:
Name                               Description
MongoValidatorSelecter             Class to aid in the selection of a validator for a mongo document field

Exceptions:
None

Functions:
None


Copyright (c) 2009 Christopher Perkins
Original Version by Christopher Perkins 2009
Released under MIT license.

Mongo Contributions by Jorge Vargas
"""

from sprox.validatorselector import ValidatorSelector

class MongoKitValidatorSelector(ValidatorSelector):
    pass




from tw.api import Widget
from tw.forms import CalendarDatePicker, CalendarDateTimePicker, TableForm, DataGrid
from tw.forms.fields import (SingleSelectField, MultipleSelectField, InputField, HiddenField,
                             TextField, FileField, PasswordField, TextArea, Label)

from formencode.schema import Schema
from formencode.validators import StringBool
from formencode import Invalid

class SproxMethodPutHiddenField(HiddenField):
    available_engines = ['mako', 'genshi']
    template="sprox.widgets.tw1widgets.templates.hidden_put"

class SproxCalendarDatePicker(CalendarDatePicker):
    date_format = '%Y-%m-%d'

class SproxTimePicker(CalendarDateTimePicker):
    date_format = '%H:%M:%S'

class SproxCalendarDateTimePicker(CalendarDateTimePicker):
    date_format = '%Y-%m-%d %H:%M:%S'

class SproxDataGrid(DataGrid):
    available_engines = ['mako', 'genshi']
    template = "sprox.widgets.tw1widgets.templates.datagrid"
    params = ['pks', 'controller', 'xml_fields']
    xml_fields = ['actions']

class ContainerWidget(Widget):
    template = "genshi:sprox.widgets.tw1widgets.templates.container"
    params = ["controller",]

class TableLabelWidget(Widget):
    template = "genshi:sprox.widgets.tw1widgets.templates.tableLabel"
    params = ["identifier", "controller"]

class ModelLabelWidget(Widget):
    available_engines = ['mako', 'genshi']
    template = "sprox.widgets.tw1widgets.templates.modelLabel"
    params = ["identifier", "controller"]

class EntityLabelWidget(Widget):
    template = "genshi:sprox.widgets.tw1widgets.templates.entityLabel"
    params = ["entity", "controller"]

class RecordViewWidget(Widget):
    template = "genshi:sprox.widgets.tw1widgets.templates.recordViewTable"
    params = ["entity"]

class RecordFieldWidget(Widget):
    template = "genshi:sprox.widgets.tw1widgets.templates.recordField"
    params = ['field_name']

class TableDefWidget(Widget):
    template = "genshi:sprox.widgets.tw1widgets.templates.tableDef"
    params = ["identifier"]

class EntityDefWidget(Widget):
    available_engines = ['genshi']
    template = "sprox.widgets.tw1widgets.templates.entityDef"
    params = ["entity"]

class TableWidget(Widget):
    available_engines = ['genshi']
    template = "genshi:sprox.widgets.tw1widgets.templates.table"

class SproxTableForm(TableForm):
    available_engines = ['mako', 'genshi']
    validator = Schema(ignore_missing_keys=True, allow_extra_fields=True)
    template = "sprox.widgets.tw1widgets.templates.tableForm"

#custom checkbox widget since I am not happy with the behavior of the TW one
class SproxCheckBox(InputField):
    available_engines = ['mako', 'genshi']
    template = "sprox.widgets.tw1widgets.templates.checkbox"
    validator = StringBool
    def update_params(self, d):
        InputField.update_params(self, d)
        try:
            checked = self.validator.to_python(d.value)
        except Invalid:
            checked = False
        d.attrs['checked'] = checked or None

class PropertyMixin(Widget):
    params = ['entity', 'field_name', 'provider', 'dropdown_field_names']

    def _my_update_params(self, d, nullable=False):
        entity = self.entity
        options = self.provider.get_dropdown_options(self.entity, self.field_name, self.dropdown_field_names)
        if nullable:
            options.append([None,"-----------"])
        if len(options) == 0:
            return {}
        d['options']= options

        return d

class PropertySingleSelectField(SingleSelectField, PropertyMixin):
    params=["nullable", "disabled"]
    nullable=False
    disabled=False
    def update_params(self, d):
        self._my_update_params(d,nullable=self.nullable)
        SingleSelectField.update_params(self, d)
        return d

class PropertyMultipleSelectField(MultipleSelectField, PropertyMixin):
    params=["disabled"]
    disabled=False
    def update_params(self, d):
        self._my_update_params(d)
        MultipleSelectField.update_params(self, d)
        return d




"""
fillerbase Module

Classes to help fill widgets with data

Copyright (c) 2008 Christopher Perkins
Original Version by Christopher Perkins 2008
Released under MIT license.
"""

from sprox.formbase import FormBase, EditableForm, AddRecordForm
from sprox.widgets.tw1widgets.dojo import SproxDojoSelectShuttleField, SproxDojoSortedSelectShuttleField

SAWidgetSelector = None
SAORMProvider = None
try:
    from sprox.sa.widgetselector import SAWidgetSelector
    from sprox.sa.provider import SAORMProvider
except ImportError: # pragma: no cover
    pass

class DojoSAWidgetSelector(SAWidgetSelector):
    """Dojo-Specific Widget Selector"""
    default_multiple_select_field_widget_type = SproxDojoSelectShuttleField

class DojoFormBase(FormBase):
    """FormBase for Dojo

    see :class:`sprox.formbase.FormBase`

    """
    @property
    def __widget_selector_type__(self):
        if isinstance(self.__provider__, SAORMProvider):
            return DojoSAWidgetSelector
        return super(DojoFormBase, self).__widget_selector_type__

class DojoEditableForm(EditableForm):
    """Creates a form for editing records that has select shuttles for the multiple relations.

    :Modifiers:
      see :class:`sprox.formbase.FormBase`

    :Usage:

    >>> from sprox.dojo.formbase import DojoEditableForm
    >>> from formencode import Schema
    >>> from formencode.validators import FieldsMatch
    >>> class Form(DojoEditableForm):
    ...     __model__ = User
    ...     __limit_fields__       = ['user_name', 'groups']
    >>> edit_form = Form()
    >>> print edit_form() # doctest: +XML
    <form action="" method="post" class="required tableform">
        <div>
                <input type="hidden" id="sprox_id" class="hiddenfield" name="sprox_id" value="" />
                <input type="hidden" id="user_id" class="hiddenfield" name="user_id" value="" />
                <input type="hidden" name="_method" class="sproxmethodputhiddenfield" id="sprox_method" value="PUT" />
        </div>
        <table border="0" cellspacing="0" cellpadding="2" >
            <tr class="even" id="user_name.container" title="" >
                <td class="labelcol">
                    <label id="user_name.label" for="user_name" class="fieldlabel">User Name</label>
                </td>
                <td class="fieldcol" >
                    <input type="text" id="user_name" class="textfield" name="user_name" value="" />
                </td>
            </tr>
            <tr class="odd" id="groups.container" title="" >
                <td class="labelcol">
                    <label id="groups.label" for="groups" class="fieldlabel">Groups</label>
                </td>
                <td class="fieldcol" >
                    <div xmlns="http://www.w3.org/1999/xhtml" dojoType="twdojo.SelectShuttle" id="groups_SelectShuttle">
        <div style="float:left; padding: 5px; width:10em;">
            Available<br />
            <select class="shuttle" id="groups_src" multiple="multiple" name="" size="5">
                    <option value="1">0</option><option value="2">1</option><option value="3">2</option><option value="4">3</option><option value="5">4</option>
            </select>
        </div>
        <div style="float:left; padding: 25px 5px 5px 0px;" id="groups_Buttons">
            <button class="shuttle" id="groups_AllRightButton">&gt;&gt;</button><br />
            <button class="shuttle" id="groups_RightButton">&gt;</button><br />
            <button class="shuttle" id="groups_LeftButton">&lt;</button><br />
            <button class="shuttle" id="groups_AllLeftButton">&lt;&lt;</button>
        </div>
        <div style="float:left; padding: 5px; width:10em;">
                Selected<br />
                <select class="shuttle" id="groups" multiple="multiple" name="groups" size="5">
                </select>
        </div>
        <script type="text/javascript">
        //create an object of this type here
        </script>
    </div>
                </td>
            </tr>
            <tr class="even" id="user_id.container" title="" >
                <td class="labelcol">
                    <label id="user_id.label" for="user_id" class="fieldlabel required">User</label>
                </td>
                <td class="fieldcol" >
                    <input type="text" id="user_id" class="textfield required" name="user_id" value="" disabled="disabled" />
                </td>
            </tr>
            <tr class="odd" id="submit.container" title="" >
                <td class="labelcol">
                    <label id="submit.label" for="submit" class="fieldlabel"></label>
                </td>
                <td class="fieldcol" >
                    <input type="submit" class="submitbutton" value="Submit" />
                </td>
            </tr>
        </table>
    </form>
"""
    @property
    def __widget_selector_type__(self):
        if isinstance(self.__provider__, SAORMProvider):
            return DojoSAWidgetSelector
        return super(EditableForm, self).__widget_selector_type__

class DojoAddRecordForm(AddRecordForm):
    """
    Creates a form for adding records that has select shuttles for the multiple relations.

    :Modifiers:
      see :class:`sprox.formbase.FormBase`

    :Usage:

    >>> from sprox.dojo.formbase import DojoAddRecordForm
    >>> from formencode import Schema
    >>> from formencode.validators import FieldsMatch
    >>> class Form(DojoAddRecordForm):
    ...     __model__ = User
    ...     __limit_fields__       = ['user_name', 'groups']
    >>> add_form = Form()
    >>> print add_form() # doctest: +XML
    <form action="" method="post" class="required tableform">
        <div>
                <input type="hidden" id="sprox_id" class="hiddenfield" name="sprox_id" value="" />
        </div>
        <table border="0" cellspacing="0" cellpadding="2" >
            <tr class="even" id="user_name.container" title="" >
                <td class="labelcol">
                    <label id="user_name.label" for="user_name" class="fieldlabel">User Name</label>
                </td>
                <td class="fieldcol" >
                    <input type="text" id="user_name" class="textfield" name="user_name" value="" />
                </td>
            </tr>
            <tr class="odd" id="groups.container" title="" >
                <td class="labelcol">
                    <label id="groups.label" for="groups" class="fieldlabel">Groups</label>
                </td>
                <td class="fieldcol" >
                    <div xmlns="http://www.w3.org/1999/xhtml" dojoType="twdojo.SelectShuttle" id="groups_SelectShuttle">
        <div style="float:left; padding: 5px; width:10em;">
            Available<br />
            <select class="shuttle" id="groups_src" multiple="multiple" name="" size="5">
                    <option value="1">0</option><option value="2">1</option><option value="3">2</option><option value="4">3</option><option value="5">4</option>
            </select>
        </div>
        <div style="float:left; padding: 25px 5px 5px 0px;" id="groups_Buttons">
            <button class="shuttle" id="groups_AllRightButton">&gt;&gt;</button><br />
            <button class="shuttle" id="groups_RightButton">&gt;</button><br />
            <button class="shuttle" id="groups_LeftButton">&lt;</button><br />
            <button class="shuttle" id="groups_AllLeftButton">&lt;&lt;</button>
        </div>
        <div style="float:left; padding: 5px; width:10em;">
                Selected<br />
                <select class="shuttle" id="groups" multiple="multiple" name="groups" size="5">
                </select>
        </div>
        <script type="text/javascript">
        //create an object of this type here
        </script>
    </div>
                </td>
            </tr>
            <tr class="even" id="submit.container" title="" >
                <td class="labelcol">
                    <label id="submit.label" for="submit" class="fieldlabel"></label>
                </td>
                <td class="fieldcol" >
                    <input type="submit" class="submitbutton" value="Submit" />
                </td>
            </tr>
        </table>
    </form>
"""
    @property
    def __widget_selector_type__(self):
        if isinstance(self.__provider__, SAORMProvider):
            return DojoSAWidgetSelector
        return super(AddRecordForm, self).__widget_selector_type__



"""
util Module

this contains the class which allows dbsprockets to interface with sqlalchemy.

Copyright (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""

from copy import deepcopy, copy

"""
A good portion of this code was lifted from the PyYaml Codebase.

http://pyyaml.org/:
Copyright (c) 2006 Kirill Simonov

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re, datetime

class ConverterError(Exception):pass

timestamp_regexp = re.compile(
        r'''^(?P<year>[0-9][0-9][0-9][0-9])
            -(?P<month>[0-9][0-9]?)
            -(?P<day>[0-9][0-9]?)
            (?:(?:[Tt]|[ \t]+)
            (?P<hour>[0-9][0-9]?)
            :(?P<minute>[0-9][0-9])
            :(?P<second>[0-9][0-9])
            (?:\.(?P<fraction>[0-9]*))?
            (?:[ \t]*(?P<tz>Z|(?P<tz_sign>[-+])(?P<tz_hour>[0-9][0-9]?)
            (?::(?P<tz_minute>[0-9][0-9]))?))?)?$''', re.X)

def timestamp(value):
    match = timestamp_regexp.match(value)
    if match is None:
        raise ConverterError('Unknown DateTime format, %s try %%Y-%%m-%%d %%h:%%m:%%s.d'%value)
    values = match.groupdict()
    year = int(values['year'])
    month = int(values['month'])
    day = int(values['day'])
    if not values['hour']:
        return datetime.date(year, month, day)
    hour = int(values['hour'])
    minute = int(values['minute'])
    second = int(values['second'])
    fraction = 0
    if values['fraction']:
        fraction = values['fraction'][:6]
        while len(fraction) < 6:
            fraction += '0'
        fraction = int(fraction)
    delta = None
    if values['tz_sign']:
        tz_hour = int(values['tz_hour'])
        tz_minute = int(values['tz_minute'] or 0)
        delta = datetime.timedelta(hours=tz_hour, minutes=tz_minute)
        if values['tz_sign'] == '-':
            delta = -delta
    data = datetime.datetime(year, month, day, hour, minute, second, fraction)
    if delta:
        data -= delta
    return data

def name2label(name):
    """
    Took from ToscaWidgets2.

    Convert a column name to a Human Readable name.
       1) Strip _id from the end
       2) Convert _ to spaces
       3) Convert CamelCase to Camel Case
       4) Upcase first character of Each Word
    """
    if name.endswith('_id'):
        name = name[:-3]
    return ' '.join([s.capitalize() for s in
                     re.findall(r'([A-Z][a-z0-9]+|[a-z0-9]+|[A-Z0-9]+)', name)])

try: #pragma: no cover
    from tw2.core import Widget
    from tw2.core.widgets import WidgetMeta
    from tw2.forms import HiddenField
except ImportError: #pragma: no cover
    from tw.api import Widget
    from tw.forms import HiddenField
    class WidgetMeta(object):
        """TW2 WidgetMetaClass"""

def is_widget(w):
    if hasattr(w, 'req'):
        return isinstance(w, Widget) or isinstance(w, WidgetMeta) and w.__name__.endswith('_s')
    else:
        return isinstance(w, Widget)

def is_widget_class(w):
    if hasattr(w, 'req'):
        return isinstance(w, WidgetMeta) and not w.__name__.endswith('_s')
    else:
        return issubclass(w, Widget)



"""
tablebase Module

Classes to create table widgets.

Copyright (c) 2008-10 Christopher Perkins
Original Version by Christopher Perkins 2008
Released under MIT license.
"""
from operator import itemgetter

try: #pragma: no cover
    import tw2.forms
    from tw2.core import Widget
    from tw2.core.widgets import WidgetMeta
    from tw2.forms.datagrid import Column
except ImportError: #pragma: no cover
    from tw.forms.datagrid import Column
    from tw.api import Widget
    class WidgetMeta(object):
        pass

from sprox.widgets import SproxDataGrid
from sprox.viewbase import ViewBase
from sprox.metadata import FieldsMetadata

class TableBase(ViewBase):
    """This class allows you to create a table widget.

    :Modifiers:
      +-----------------------------------+--------------------------------------------+------------------------------+
      | Name                              | Description                                | Default                      |
      +===================================+============================================+==============================+
      | __base_widget_type__              | Base widget for fields to go into.         | SproxDataGrid                |
      +-----------------------------------+--------------------------------------------+------------------------------+
      | __metadata_type__                 | Type the widget is based on.               | FieldsMetadata               |
      +-----------------------------------+--------------------------------------------+------------------------------+
      | __headers__                       | A dictionary of field/header pairs.        | {}                           |
      +-----------------------------------+--------------------------------------------+------------------------------+
      | __column_widths__                 | A dictionary of field/width(string) pairs. | {}                           |
      +-----------------------------------+--------------------------------------------+------------------------------+
      | __default_column_width__          | Header size to use when not specified in   | '10em'                       |
      |                                   | __column_widths__                          |                              |
      +-----------------------------------+--------------------------------------------+------------------------------+
      | __xml_fields__                    | fields whose values should show as html    |                              |
      +-----------------------------------+--------------------------------------------+------------------------------+


    see modifiers in :mod:`sprox.viewbase`

    Here is an example listing of the towns in the test database.


    >>> from sprox.tablebase import TableBase
    >>> class TownTable(TableBase):
    ...    __model__ = Town
    >>> town_table = TownTable(session)
    >>> print(town_table()) #doctest: +XML
    <div>
    <table class="grid">
        <thead>
            <tr>
                    <th  class="col_0">actions</th>
                    <th  class="col_1">town_id</th>
                    <th  class="col_2">name</th>
            </tr>
        </thead>
        <tbody>
        </tbody>
    </table>
          No Records Found.
    </div>

    As you can see, this is none too interesting, because there is no data in the table.
    Here is how we fill the table with data using TableFillerBase

    >>> from sprox.fillerbase import TableFiller
    >>> class TownFiller(TableFiller):
    ...     __model__ = Town
    >>> town_filler = TownFiller(session)
    >>> value = town_filler.get_value()
    >>> print(town_table(value=value)) #doctest: +SKIP
    <div xmlns="http://www.w3.org/1999/xhtml">
    <div>
    <table
           id="None" class="grid">
        <thead>
            <tr>
                    <th class="col_0">town_id</th>
                    <th class="col_1">name</th>
            </tr>
        </thead>
        <tbody>
            <tr class="even">
                <td class="col_0">
                        1
                </td>
                <td class="col_1">
                        Arvada
                </td>
            </tr>
            <tr class="odd">
                <td class="col_0">
                        2
                </td>
                <td class="col_1">
                        Denver
                </td>
            </tr>
            <tr class="even">
                <td class="col_0">
                        3
                </td>
                <td class="col_1">
                        Golden
                </td>
            </tr>
            <tr class="odd">
                <td class="col_0">
                        4
                </td>
                <td class="col_1">
                        Boulder
                </td>
            </tr>
        </tbody>
    </table>
    </div>

    And now you can see the table has some data in it, and some restful links to the data.  But what if you don't want those links?
    You can omit the links by adding '__actions__' to omitted fields as follows:

    >>> class TownTable(TableBase):
    ...     __model__ = Town
    ...     __omit_fields__ = ['__actions__']
    >>> town_table = TownTable(session)
    >>> print(town_table(value=value)) #doctest: +XML
    <div>
    <table class="grid">
        <thead>
            <tr>
                    <th  class="col_0">town_id</th>
                    <th  class="col_1">name</th>
            </tr>
        </thead>
        <tbody>
                <tr class="even">
                    <td class="col_0">
                            1
                    </td>
                    <td class="col_1">
                            Arvada
                    </td>
                </tr>
                <tr class="odd">
                    <td class="col_0">
                            2
                    </td>
                    <td class="col_1">
                            Denver
                    </td>
                </tr>
                <tr class="even">
                    <td class="col_0">
                            3
                    </td>
                    <td class="col_1">
                            Golden
                    </td>
                </tr>
                <tr class="odd">
                    <td class="col_0">
                            4
                    </td>
                    <td class="col_1">
                            Boulder
                    </td>
                </tr>
        </tbody>
    </table>
    </div>

    """

    #object overrides
    __base_widget_type__ = SproxDataGrid
    __metadata_type__    = FieldsMetadata
    __headers__          = None
    __column_widths__    = None
    __xml_fields__       = None
    __default_column_width__ = "10em"

    def _do_get_fields(self):
        fields = super(TableBase, self)._do_get_fields()
        if '__actions__' not in self.__omit_fields__ and '__actions__' not in fields:
            fields.insert(0, '__actions__')
            if '__actions__' not in self.__headers__:
                self.__headers__['__actions__'] = 'actions'
        return fields

    def _do_init_attrs(self):
        super(TableBase, self)._do_init_attrs()
        if self.__headers__ is None:
            self.__headers__ = {}
        if self.__column_widths__ is None:
            self.__column_widths__ = {}
        if self.__xml_fields__ is None:
            self.__xml_fields__ = {}

    def _table_field_is_plain_widget(self, widget):
        if widget.__class__ == Widget or\
           (widget.__class__ == WidgetMeta and Widget in widget.__bases__):
            return True

        return False

    def _do_get_widget_args(self):
        args = super(TableBase, self)._do_get_widget_args()
        args['pks'] = None
        args['column_widths'] = self.__column_widths__
        args['default_column_width'] = self.__default_column_width__

        # Okay: the trunk version of Sprox creates its args['fields']
        # very simply, as a list of lambdas that do an item lookup on
        # the field name.  This means that TableBase inherently ignores
        # any attempt to specify widgets for displaying or formatting
        # model values; in fact, the field widget list that the
        # TableBase grows at some point during processing appears to be
        # completely ignored.  (You will see if it you add "print"
        # statements around your code to watch how your tables are
        # behaving.)

        # To make widgets active, we need to put the widgets in place of
        # the lambdas in the the args['fields'] that we build.  There
        # are challenges to building this list, however:

        # 1. The widgets supplied by default are useless.  Instead of
        #    being something that can actually display text, the list of
        #    widgets somehow winds up just being bare Widget instances,
        #    which seem to display absolutely nothing when rendered.
        #    Therefore, when we see a bare Widget supplied for a
        #    particular column, we need to ignore it.

        # 2. Some fields (like '__action__') do not even appear in our
        #    list of fields-plus-widgets called self.__fields__, so we
        #    have to be willing to fall back to a lambda in that case
        #    too.

        # 3. For some reason, the list of field widgets is not always
        #    present, so we have to wastefully re-compute it here by
        #    calling the _do_get_field_widgets() method.  It would be
        #    nice if this were produced once before these computations
        #    started and could be relied upon to be present, but it's
        #    not so far as I could tell.  But, then, I was reading code
        #    late at night.

        # Anyway, with the stanza below, I am able to provide my table
        # classes with a __field_widget_types__ dictionary and have
        # those fields rendered differently.

        field_widget_dict = self._do_get_field_widgets(self.__fields__)

        field_columns = []
        for field in self.__fields__:
            widget = field_widget_dict.get(field, None)
            if widget is None or self._table_field_is_plain_widget(widget): # yuck
                column = Column(field, itemgetter(field), self.__headers__.get(field, field))
            else:
                column = Column(field, widget, self.__headers__.get(field, field))
            field_columns.append(column)
        args['fields'] = field_columns

        # And, now back to our regularly-scheduled trunk-derived Sprox.
        if '__actions__' not in self.__omit_fields__:
            args['pks'] = self.__provider__.get_primary_fields(self.__entity__)

        args['xml_fields'] = self.__xml_fields__
        args.pop('children', None)
        return args



try: #pragma: no cover
    from .tw2widgets.widgets import *
except ImportError: #pragma: no cover
    from .tw1widgets.widgets import *

try: #pragma: no cover
    from tw2.forms import CalendarBase
except: #pragma: no cover
    class CalendarBase(object):
        pass



"""
saprovider Module

this contains the class which allows dbsprockets to interface with sqlalchemy.

Classes:
Name                               Description
SAProvider                         sqlalchemy metadata/crud provider

Exceptions:
None

Functions:
None


Copyright (c) 2007 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""
import inspect
import re
from sqlalchemy import and_, or_, DateTime, Date, Interval, Integer, MetaData, desc as _desc, func
from sqlalchemy import String
from sqlalchemy.engine import Engine
from sqlalchemy.orm.session import Session
from sqlalchemy.orm.scoping import ScopedSession
from sqlalchemy.orm.query import Query
from sqlalchemy.orm import Mapper, _mapper_registry, SynonymProperty, object_mapper, class_mapper
from sqlalchemy.orm.exc import UnmappedClassError, NoResultFound, UnmappedInstanceError
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.schema import Column
from sprox.sa.support import PropertyLoader, resolve_entity, Binary, LargeBinary

from sprox.iprovider import IProvider
from cgi import FieldStorage
from datetime import datetime, date, timedelta
from warnings import warn

from sprox.sa.widgetselector import SAWidgetSelector
from sprox.sa.validatorselector import SAValidatorSelector
from sprox._compat import string_type, zip_longest

class SAORMProviderError(Exception):pass

class SAORMProvider(IProvider):
    default_view_names = ['_name', 'name', 'description', 'title']

    default_widget_selector_type = SAWidgetSelector
    default_validator_selector_type = SAValidatorSelector

    def __init__(self, hint=None, **hints):
        """
        initialize me with a engine, bound metadata or bound session object
        or a combination of the above.
        """
        #if hint is None and len(hints) == 0:
        #    raise SAORMProviderError('You must provide a hint to this provider')
        self.engine, self.session, self.metadata = self._get_engine(hint, hints)

    def _get_engine(self, hint, hints):
        metadata = hints.get('metadata', None)
        engine   = hints.get('engine', None)
        session  = hints.get('session', None)

        if isinstance(hint, Engine):
            engine=hint

        if isinstance(hint, MetaData):
            metadata=hint

        if isinstance(hint, (Session, ScopedSession)):
            session = hint

        if session is not None and engine is None:
            engine = session.bind

        if metadata is not None and engine is None:
            engine = metadata.bind

        return engine, session, metadata

    def get_fields(self, entity):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        field_names = list(mapper.c.keys())
        for prop in mapper.iterate_properties:
            try:
                mapper.c[prop.key]
                field_names.append(prop.key)
            except KeyError:
                mapper.get_property(prop.key)
                field_names.append(prop.key)

        return field_names

    def get_entity(self, name):
        for mapper in _mapper_registry:
            if mapper.class_.__name__ == name:
                engine = mapper.tables[0].bind
                if engine is not None and mapper.tables[0].bind != self.engine:
                    continue
                return mapper.class_
        raise KeyError('could not find model by the name %s'%(name))

    def get_entities(self):
        entities = []
        for mapper in _mapper_registry:
            engine = mapper.tables[0].bind
            if engine is not None and mapper.tables[0].bind != self.engine:
                continue
            entities.append(mapper.class_.__name__)
        return entities

    def get_field(self, entity, name):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        try:
            return mapper.c[name]
        except (InvalidRequestError, KeyError):
            try:
                return mapper.get_property(name)
            except InvalidRequestError:
                raise AttributeError



    def is_binary(self, entity, name):
        field = self.get_field(entity, name)
        if isinstance(field, PropertyLoader):
            field = self._relationship_local_side(field)[0]
        if isinstance(field, SynonymProperty):
            field = self.get_field(entity, field.name)
        return isinstance(field.type, (Binary, LargeBinary))

    def is_nullable(self, entity, name):
        field = self.get_field(entity, name)
        if isinstance(field, SynonymProperty):
            field = self.get_field(entity, field.name)
        if isinstance(field, PropertyLoader):
            return getattr(self._relationship_local_side(field)[0], 'nullable')
        return getattr(field, 'nullable', True)

    def is_string(self, entity, name):
        field = self.get_field(entity, name)
        if isinstance(field, PropertyLoader):
            field = self._relationship_local_side(field)[0]
        if isinstance(field, SynonymProperty):
            field = self.get_field(entity, field.name)

        #should cover Unicode, UnicodeText, Text and so on by inheritance
        return isinstance(field.type, String)

    def get_primary_fields(self, entity):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        fields = []

        for field_name in self.get_fields(entity):
            try:
                value = mapper.c[field_name]
            except KeyError:
                # Relations won't be attributes, but can't be primary anyway.
                continue
            if value.primary_key and not field_name in fields:
                fields.append(field_name)
        return fields

    def get_primary_field(self, entity):
        fields = self.get_primary_fields(entity)
        assert len(fields) > 0
        return fields[0]

    def _find_title_column(self, entity):
        entity = resolve_entity(entity)
        for column in class_mapper(entity).columns:
            if 'title' in column.info and column.info['title']:
                return column.key
        return None

    def get_view_field_name(self, entity, possible_names, item=None):
        view_field = self._find_title_column(entity)

        fields = self.get_fields(entity)
        for column_name in possible_names:
            for actual_name in fields:
                if column_name == actual_name:
                    view_field = actual_name
                    break
            if view_field:
                break
            for actual_name in fields:
                if column_name in actual_name:
                    view_field = actual_name
                    break
            if view_field:
                break
        if view_field is None:
            view_field = fields[0]
        return view_field

    def get_dropdown_options(self, entity, field_name, view_names=None):
        if view_names is None:
            view_names = self.default_view_names

        if self.session is None:
            warn('No dropdown options will be shown for %s.  '
                 'Try passing the session into the initialization '
                 'of your form base object so that this sprocket '
                 'can have values in the drop downs'%entity)
            return []

        field = self.get_field(entity, field_name)

        target_field = entity
        if isinstance(field, PropertyLoader):
            target_field = field.argument
        target_field = resolve_entity(target_field)

        #some kind of relation
        if isinstance(target_field, Mapper):
            target_field = target_field.class_

        pk_fields = self.get_primary_fields(target_field)

        view_name = self.get_view_field_name(target_field, view_names)

        rows = self.session.query(target_field).all()

        if len(pk_fields) == 1:
            def build_pk(row):
                return getattr(row, pk_fields[0])
        else:
            def build_pk(row):
                return "/".join([str(getattr(row, pk)) for pk in pk_fields])

        return [ (build_pk(row), getattr(row, view_name)) for row in rows ]

    def get_relations(self, entity):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        return [prop.key for prop in mapper.iterate_properties if isinstance(prop, PropertyLoader)]

    def is_relation(self, entity, field_name):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)

        try:
            property = mapper.get_property(field_name)
        except InvalidRequestError:
            return False

        if isinstance(property, PropertyLoader):
            return True

    def _relates_many(self, entity, field_name):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        property = mapper.get_property(field_name)
        return property.uselist

    def relation_fields(self, entity, field_name):
        field = getattr(entity, field_name)
        return [col.name for col in self._relationship_local_side(field.property)]

    def is_query(self, entity, value):
        """determines if a field is a query instead of actual list of data"""
        return isinstance(value, Query)

    def is_unique(self, entity, field_name, value):
        entity = resolve_entity(entity)
        field = getattr(entity, field_name)
        try:
            self.session.query(entity).filter(field==value).one()
        except NoResultFound:
            return True
        return False

    def get_synonyms(self, entity):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        return [prop.key for prop in mapper.iterate_properties if isinstance(prop, SynonymProperty)]

    def _adapt_type(self, value, primary_key):
        if isinstance(primary_key.type, Integer):
            value = int(value)
        return value

    def _relationship_local_side(self, relationship):
        if hasattr(relationship, 'local_columns'):
            return list(relationship.local_columns)
        return list(relationship.local_side) #pragma: no cover

    def _modify_params_for_relationships(self, entity, params):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        relations = self.get_relations(entity)

        for relation in relations:
            if relation in params:
                prop = mapper.get_property(relation)
                target = prop.argument
                target = resolve_entity(target)
                value = params[relation]
                if value:
                    if prop.uselist and isinstance(value, list):
                        target_obj = []
                        for v in value:
                            try:
                                object_mapper(v)
                                target_obj.append(v)
                            except UnmappedInstanceError:
                                if hasattr(target, 'primary_key'):
                                    pk = target.primary_key
                                else:
                                    pk = class_mapper(target).primary_key
                                if isinstance(v, string_type) and "/" in v:
                                    v = map(self._adapt_type, v.split("/"), pk)
                                    v = tuple(v)
                                else:
                                    v = self._adapt_type(v, pk[0])
                                #only add those items that come back
                                new_v = self.session.query(target).get(v)
                                if new_v is not None:
                                    target_obj.append(new_v)
                    elif prop.uselist:
                        try:
                            object_mapper(value)
                            target_obj = [value]
                        except UnmappedInstanceError:
                            mapper = target
                            if not isinstance(target, Mapper):
                                mapper = class_mapper(target)
                            value = self._adapt_type(value, mapper.primary_key[0])
                            target_obj = [self.session.query(target).get(value)]
                    else:
                        try:
                            object_mapper(value)
                            target_obj = value
                        except UnmappedInstanceError:
                            if isinstance(value, string_type) and "/" in value:
                                value = map(self._adapt_type, value.split("/"), prop.remote_side)
                                value = tuple(value)
                            else:
                                value = self._adapt_type(value, list(prop.remote_side)[0])
                            target_obj = self.session.query(target).get(value)
                    params[relation] = target_obj
                else:
                    if prop.uselist:
                        params[relation] = []
                    else:
                        params[relation] = None

        return params

    def create(self, entity, params):
        entity = resolve_entity(entity)
        params = self._modify_params_for_dates(entity, params)
        params = self._modify_params_for_relationships(entity, params)
        obj = entity()

        relations = self.get_relations(entity)
        mapper = class_mapper(entity)
        for key, value in params.items():
            if value is not None:
                if isinstance(value, FieldStorage):
                    value = value.file.read()
                try:
                    if key not in relations and value and isinstance(mapper.columns[key].type, Integer):
                        value = int(value)
                except KeyError:
                    pass
                setattr(obj, key, value)

        self.session.add(obj)
        self.session.flush()
        return obj

    def flush(self):
        self.session.flush()

    def dictify(self, obj, fields=None, omit_fields=None):
        if obj is None:
            return {}
        r = {}
        mapper = class_mapper(obj.__class__)
        for prop in mapper.iterate_properties:
            if fields and prop.key not in fields:
                continue

            if omit_fields and prop.key in omit_fields:
                continue

            value = getattr(obj, prop.key)
            if value is not None:
                if isinstance(prop, PropertyLoader):
                    klass = prop.argument
                    if isinstance(klass, Mapper):
                        klass = klass.class_
                    pk_name = self.get_primary_field(klass)
                    if isinstance(value, list) or isinstance(value, Query):
                        #joins
                        value = [getattr(value, pk_name) for value in value]
                    else:
                        #fks
                        value = getattr(value, pk_name)
            r[prop.key] = value
        return r

    def get_field_default(self, field):
        if isinstance(field, Column) and field.default and getattr(field.default, 'arg', None) is not None:
            if isinstance(field.default.arg, (string_type, int, float)):
                return (True, field.default.arg)
            elif callable(field.default.arg):
                # SQLAlachemy wraps default so that they receive a context, but TW can't provide one
                return (True, lambda: field.default.arg(None))
        return (False, None)

    def get_field_provider_specific_widget_args(self, entity, field, field_name):
        args = {}
        if isinstance(field, PropertyLoader):
            args['provider'] = self
            args['nullable'] = self.is_nullable(entity, field_name)
        return args

    def get_default_values(self, entity, params):
        return params

    def get_obj(self, entity, params, fields=None, omit_fields=None):
        obj = self._get_obj(entity, params)
        return obj

    def get(self, entity, params, fields=None, omit_fields=None):
        obj = self.get_obj(entity, params, fields)
        return self.dictify(obj, fields, omit_fields)

    def _escape_like(self, value):
        return value.replace('*', '**').replace('%', '*%').replace('_', '*_')

    def _get_related_class(self, entity, relation):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        prop = mapper.get_property(relation)

        target = resolve_entity(prop.argument)
        if not hasattr(target, 'class_'):
            target = class_mapper(target)

        return target.class_

    def _modify_params_for_related_searches(self, entity, params, view_names=None, substrings=()):
        if view_names is None:
            view_names = self.default_view_names

        relations = self.get_relations(entity)
        for relation in relations:
            if relation in params:
                value = params[relation]
                if not isinstance(value, string_type):
                    # When not a string consider it the related class primary key
                    params.update(self._modify_params_for_relationships(entity, {relation: value}))
                    continue

                if not value:
                    # As we use ``contains``, searching for an empty text
                    # will just lead to all results so we just remove the filter.
                    del params[relation]
                    continue

                target_class = self._get_related_class(entity, relation)
                view_name = self.get_view_field_name(target_class, view_names)
                related_column = getattr(target_class, view_name)

                if relation in substrings:
                    escaped_value = self._escape_like(value.lower())
                    filter = func.lower(related_column).contains(
                        escaped_value, escape='*'
                    )
                else:
                    filter = (func.lower(related_column) == value.lower())

                params[relation] = self.session.query(target_class).filter(filter).all()

        return params

    def query(self, entity, limit=None, offset=None, limit_fields=None,
              order_by=None, desc=False, field_names=[], filters={},
              substring_filters=[], search_related=False, related_field_names=None,
              **kw):
        entity = resolve_entity(entity)
        query = self.session.query(entity)

        filters = self._modify_params_for_dates(entity, filters)

        if search_related:
            # Values for related fields contain the text to search
            filters = self._modify_params_for_related_searches(entity, filters,
                                                               view_names=related_field_names,
                                                               substrings=substring_filters)
        else:
            # Values for related fields contain the primary key
            filters = self._modify_params_for_relationships(entity, filters)

        for field_name, value in filters.items():
            field = getattr(entity, field_name)
            if self.is_relation(entity, field_name) and isinstance(value, list):
                related_class = self._get_related_class(entity, field_name)
                related_pk = self.get_primary_field(related_class)
                related_pk_col = getattr(related_class, related_pk)
                related_pk_values = (getattr(v, related_pk) for v in value)
                if self._relates_many(entity, field_name):
                    field_filter = field.any(related_pk_col.in_(related_pk_values))
                else:
                    field_filter = field.has(related_pk_col.in_(related_pk_values))
                query = query.filter(field_filter)
            elif field_name in substring_filters and self.is_string(entity, field_name):
                escaped_value = self._escape_like(value.lower())
                query = query.filter(func.lower(field).contains(escaped_value, escape='*'))
            else:
                query = query.filter(field==value)

        count = query.count()

        if order_by is not None:
            if not isinstance(order_by, (tuple, list)):
                order_by = [order_by]

            if not isinstance(desc, (tuple, list)):
                desc = [desc]

            for sort_by, sort_descending in zip_longest(order_by, desc):
                if self.is_relation(entity, sort_by):
                    mapper = class_mapper(entity)
                    class_ = None
                    for prop in mapper.iterate_properties:
                        try:
                            class_ = prop.mapper.class_
                        except (AttributeError, KeyError):
                            pass
                    query = query.join(sort_by)
                    f = self.get_view_field_name(class_, field_names)
                    field = self.get_field(class_, f)
                else:
                    field = self.get_field(entity, sort_by)

                if sort_descending:
                    field = _desc(field)
                query = query.order_by(field)

        if offset is not None:
            query = query.offset(offset)
        if limit is not None:
            query = query.limit(limit)

        objs = query.all()

        return count, objs


    def _modify_params_for_dates(self, entity, params):
        entity = resolve_entity(entity)
        mapper = class_mapper(entity)
        for key, value in list(params.items()):
            if key in mapper.c and value is not None:
                field = mapper.c[key]
                if hasattr(field, 'type'):
                    if isinstance(field.type, DateTime):
                        if not isinstance(value, datetime):
                            dt = datetime.strptime(value[:19], '%Y-%m-%d %H:%M:%S')
                            params[key] = dt
                    elif isinstance(field.type, Date):
                        if not isinstance(value, date):
                            dt = datetime.strptime(value, '%Y-%m-%d').date()
                            params[key] = dt
                    elif isinstance(field.type, Interval):
                        if not isinstance(value, timedelta):
                            d = re.match(
                                r'((?P<days>\d+) days, )?(?P<hours>\d+):'
                                r'(?P<minutes>\d+):(?P<seconds>\d+)',
                                str(value)).groupdict(0)
                            dt = timedelta(**dict(( (key, int(value))
                                                    for key, value in list(d.items()) )))
                            params[key] = dt
        return params

    def _remove_related_empty_params(self, obj, params, omit_fields=None):
        entity = obj.__class__
        mapper = class_mapper(entity)
        relations = self.get_relations(entity)
        for relation in relations:
            if omit_fields and relation in omit_fields:
                continue

            #clear out those items which are not found in the params list.
            if relation not in params or not params[relation]:
                related_items = getattr(obj, relation)
                if related_items is not None:
                    if hasattr(related_items, '__iter__'):
                        setattr(obj, relation, [])
                    else:
                        setattr(obj, relation, None)

    def _get_obj(self, entity, pkdict):
        entity = resolve_entity(entity)
        pk_names = self.get_primary_fields(entity)
        pks = tuple([pkdict[n] for n in pk_names])
        return self.session.query(entity).get(pks)

    def update(self, entity, params, omit_fields=None):
        params = self._modify_params_for_dates(entity, params)
        params = self._modify_params_for_relationships(entity, params)
        obj = self._get_obj(entity, params)
        relations = self.get_relations(entity)
        mapper = object_mapper(obj)
        for key, value in params.items():
            if omit_fields and key in omit_fields:
                continue

            if isinstance(value, FieldStorage):
                value = value.file.read()

            # this is done to cast any integer columns into ints before they are
            # sent off to the interpreter.  Oracle really needs this.
            try:
                if key not in relations and value:
                    value = self._adapt_type(value, mapper.columns[key])
            except KeyError:
                pass
            setattr(obj, key, value)

        self._remove_related_empty_params(obj, params, omit_fields)
        self.session.flush()
        return obj

    #this is hard to test because of some kind of rollback issue in the test framework
    def delete(self, entity, params):
        obj = self._get_obj(entity, params)  # pragma: no cover
        self.session.delete(obj)  # pragma: no cover
        return obj  # pragma: no cover

    def get_field_widget_args(self, entity, field_name, field):
        args = {}
        prop = getattr(field, 'property', None)
        if prop and isinstance(prop, PropertyLoader):
            args['provider'] = self
            args['nullable'] = self.is_nullable(entity, field_name)
        return args

    def is_unique_field(self, entity, field_name):
        field = self.get_field(entity, field_name)
        if hasattr(field, 'unique') and field.unique:
            return True
        return False



import inspect

try:
    from sqlalchemy.orm import PropertyLoader
except ImportError:
    # Compatibility with SQLA0.9
    from sqlalchemy.orm import RelationshipProperty as PropertyLoader

try:
    from sqlalchemy.ext.declarative.clsregistry import _class_resolver
except ImportError as e:  # pragma: no cover
    # Compatibility with SQLA < 0.9
    _class_resolver = None

try:
    # Not available in some SQLA versions
    from sqlalchemy.types import LargeBinary
except ImportError:  # pragma: no cover
    class LargeBinary:
        pass

try:
    # Future proof as it will probably be removed as deprecated
    from sqlalchemy.types import Binary
except ImportError:  # pragma: no cover
    class Binary(LargeBinary):
        pass


def resolve_entity(entity):
    if inspect.isfunction(entity):
        entity = entity()
    if _class_resolver is not None and isinstance(entity, _class_resolver):
        entity = entity()
    return entity




"""
validators Module

Copyright (c) 2008 Christopher Perkins
Original Version by Christopher Perkins 2008
Released under MIT license.
"""
from formencode import FancyValidator, Invalid
from formencode.validators import UnicodeString as FEUnicodeString
from formencode.validators import NotEmpty


class UnicodeString(FEUnicodeString):
    outputEncoding = None


class UniqueValue(FancyValidator):
    def __init__(self, provider, entity, field_name, *args, **kw):
        self.provider = provider
        self.entity   = entity
        self.field_name    = field_name
        FancyValidator.__init__(self, *args, **kw)

    def _to_python(self, value, state):
        if not self.provider.is_unique(self.entity, self.field_name, value):
            raise Invalid(
                'That value already exists',
                value, state)
        return value




"""
fillerbase Module

Classes to help fill widgets with data

Copyright (c) 2008-10 Christopher Perkins
Original Version by Christopher Perkins 2008
Released under MIT license.
"""

from .configbase import ConfigBase, ConfigBaseError
from .metadata import FieldsMetadata
import inspect
from datetime import datetime
from sprox._compat import string_type, byte_string, unicode_text
from markupsafe import Markup

encoding = 'utf-8'

class FillerBase(ConfigBase):
    """
    :Modifiers:

    see :mod:`sprox.configbase`.

    The base filler class.

    :Arguments:
      values
        pass through of values.  This is typically a set of default values that is updated by the
        filler.  This is useful when updating an existing form.
      kw
        Set of keyword arguments for assisting the fill.  This is for instance information like offset
        and limit for a TableFiller.

    :Usage:

    >>> filler = FillerBase()
    >>> filler.get_value()
    {}
    """

    def get_value(self, values=None, **kw):
        """
        The main function for getting data to fill widgets,
        """
        if values is None:
            values = {}
        return values

class ModelsFiller(FillerBase):
    pass
class ModelDefFiller(FillerBase):
    pass

class FormFiller(FillerBase):
    __metadata_type__ = FieldsMetadata

    def get_value(self, values=None, **kw):
        values = super(FormFiller, self).get_value(values)
        values['sprox_id'] =  self.__sprox_id__
        return values

class TableFiller(FillerBase):
    """
    This is the base class for generating table data for use in table widgets.  The TableFiller uses
    it's provider to obtain a dictionary of information about the __entity__ this Filler defines.
    This class is especially useful when you need to return a json stream, because it allows for
    customization of attributes.  A package which has similar functionality to this is TurboJson,
    but TurboJson is rules-based, where the semantics for generating dictionaries follows the same
    :mod:`sprox.configbase` methodology.

    Modifiers defined in this class

    +-----------------------------------+--------------------------------------------+------------------------------+
    | Name                              | Description                                | Default                      |
    +===================================+============================================+==============================+
    | __actions__                       | An overridable function to define how to   | a function that creates an   |
    |                                   | display action links in the form.          | edit and delete link.        |
    +-----------------------------------+--------------------------------------------+------------------------------+
    | __metadata_type__                 | How should we get data from the provider.  | FieldsMetadata               |
    +-----------------------------------+--------------------------------------------+------------------------------+
    | __possible_field_names__          | list or dict of names to use for discovery | None                         |
    |                                   | of field names for relationship columns.   | (See below.)                 |
    |                                   | (None uses the default list from           |                              |
    |                                   | :class:`sprox.configbase:ConfigBase`.)     |                              |
    |                                   | A dict provides field-level granularity    |                              |
    |                                   | (See also explanation below.)              |                              |
    +-----------------------------------+--------------------------------------------+------------------------------+
    | __datetime_formatstr__            | format string for the strftime function of | '%Y-%m-%d %H:%M:%S'          |
    |                                   | datetime objects.                          | ("simplified" ISO-8601)      |
    |                                   | Classical american format would be         |                              |
    |                                   | '%m/%d/%Y %H:%M%p'.                        |                              |
    +-----------------------------------+--------------------------------------------+------------------------------+

    see modifiers also in :mod:`sprox.configbase`.

    :Relations:

    By default, TableFiller will populate relations (join or foreign_key) with either the value
    from the related table, or a comma-separated list of values.  These values are derived from
    the related object given the field names provided by the __possible_field_names__ modifier.
    For instance, if you have a User class which is related to Groups, the groups item in the result
    dictionaries will be populated with Group.group_name.  The default field names are specified
    in __possible_field_name_defaults__: _name, name, description, title.

    :RESTful Actions:

    By default, Table filler provides an "__actions__" item in the resultant dictionary list.  This provides
    and edit, and (javascript) delete link which provide edit and DELETE functionality as HTML verbs in REST.
    For more information on developing RESTful URLs, please visit `http://microformats.org/wiki/rest/urls <http://microformats.org/wiki/rest/urls/>`_ .

    :Usage:

    Here is how we would get the values to fill up a user's table, minus the action column, and created date.

    >>> class UsersFiller(TableFiller):
    ...     __model__ = User
    ...     __actions__ = False
    ...     __omit_fields__ = ['created']
    >>> users_filler = UsersFiller(session)
    >>> value = users_filler.get_value(values={}, limit=20, offset=0)
    >>> print value #doctest: +IGNORE_WHITESPACE
    [{'town': u'Arvada', 'user_id': u'1', 'user_name': u'asdf',
    'town_id': u'1', 'groups': u'4', '_password': '******', 'password': '******',
    'email_address': u'asdf@asdf.com', 'display_name': u'None'}]
    """
    __actions__ = True
    __metadata_type__ = FieldsMetadata
    __possible_field_names__ = None
    __datetime_formatstr__ = '%Y-%m-%d %H:%M:%S'

    def _do_init_attrs(self):
        super(TableFiller, self)._do_init_attrs()
        if self.__possible_field_names__ is None:
            self.__possible_field_names__ = self.__possible_field_name_defaults__

    def _get_list_data_value(self, field, values):
        l = []

        if isinstance(self.__possible_field_names__, dict) and field in self.__possible_field_names__:
            view_names = self.__possible_field_names__[field]
            if not isinstance(view_names, list):
                view_names = [view_names]
        elif isinstance(self.__possible_field_names__, list):
            view_names = self.__possible_field_names__
        else:
            view_names = self.__possible_field_name_defaults__

        for value in values:
            if not isinstance(value, string_type):
                name = self.__provider__.get_view_field_name(value.__class__, view_names, value)
                l.append(unicode_text(getattr(value, name)))
            else:
                #this is needed for postgres to see array values
                return values
        return ', '.join(l)

    def _get_relation_value(self, field, value):
        #this may be needed for catwalk, but I am not sure what conditions cause it to be needed
        #if value is None:
        #    return None
        name = self.__provider__.get_view_field_name(value.__class__, self.__possible_field_names__)
        return getattr(value, name)

    def get_count(self):
        """Returns the total number of items possible for retrieval.  This can only be
        executed after a get_value() call.  This call is useful for creating pagination in the context
        of a user interface.
        """
        if not hasattr(self, '__count__'):
            raise ConfigBaseError('Count not yet set for filler.  try calling get_value() first.')
        return self.__count__

    def _do_get_fields(self):
        fields = super(TableFiller, self)._do_get_fields()
        if '__actions__' not in self.__omit_fields__ and '__actions__' not in fields:
            fields.insert(0, '__actions__')
        return fields


    def __actions__(self, obj):
        """Override this function to define how action links should be displayed for the given record."""
        primary_fields = self.__provider__.get_primary_fields(self.__entity__)
        pklist = '/'.join(map(lambda x: str(getattr(obj, x)), primary_fields))
        value = '<div><div>&nbsp;<a href="'+pklist+'/edit" style="text-decoration:none">edit</a>'\
              '</div><div>'\
              '<form method="POST" action="'+pklist+'" class="button-to">'\
            '<input type="hidden" name="_method" value="DELETE" />'\
            '<input class="delete-button" onclick="return confirm(\'Are you sure?\');" value="delete" type="submit" '\
            'style="background-color: transparent; float:left; border:0; color: #286571; display: inline; margin: 0; padding: 0;"/>'\
        '</form>'\
        '</div></div>'
        return value

    def _do_get_provider_count_and_objs(self, **kw):
        limit = kw.pop('limit', None)
        offset = kw.pop('offset', None)
        order_by = kw.pop('order_by', None)
        desc = kw.pop('desc', False)
        substring_filters = kw.pop('substring_filters', [])
        count, objs = self.__provider__.query(self.__entity__, limit, offset, self.__limit_fields__,
                                              order_by, desc, substring_filters=substring_filters,
                                              filters=kw, search_related=True,
                                              related_field_names=self.__possible_field_names__)
        self.__count__ = count
        return count, objs

    def get_value(self, values=None, **kw):
        """
        Get the values to fill a form widget.

        :Arguments:
         offset
          offset into the records
         limit
          number of records to return
         order_by
          name or list of names of the columns to the return values ordered by
         desc
          order the columns in descending order

        All the other arguments will be used to filter the result
        """
        count, objs = self._do_get_provider_count_and_objs(**kw)
        self.__count__ = count
        rows = []
        for obj in objs:
            row = {}
            for field in self.__fields__:
                field_method = getattr(self, field, None)
                if inspect.ismethod(field_method):
                    argspec = inspect.getargspec(field_method)
                    if argspec and (len(argspec[0])-2>=len(kw) or argspec[2]):
                        value = getattr(self, field)(obj, **kw)
                    else:
                        value = getattr(self, field)(obj)
                else:
                    value = getattr(obj, field)
                    if 'password' in field.lower():
                        row[field] = '******'
                        continue
                    elif isinstance(value, list) or self.__provider__.is_query(self.__entity__, value):
                        value = self._get_list_data_value(field, value)
                    elif isinstance(value, datetime):
                        value = value.strftime(self.__datetime_formatstr__)
                    elif self.__provider__.is_relation(self.__entity__, field) and value is not None:
                        value = self._get_relation_value(field, value)
                    elif self.__provider__.is_binary(self.__entity__, field) and value is not None:
                        value = Markup('&lt;file&gt;')
                if isinstance(value, byte_string):
                    value = unicode_text(value, encoding='utf-8')
                row[field] = unicode_text(value)
            rows.append(row)
        return rows

class EditFormFiller(FormFiller):
    """
    This class will help to return a single record for use within a form or otherwise.
    The values are returned in dictionary form.

    :Modifiers:

    see :mod:`sprox.configbase`.


    :Usage:

    >>> class UserFiller(EditFormFiller):
    ...     __model__ = User
    >>> users_filler = UsersFiller(session)
    >>> value = users_filler.get_value(values={'user_id':'1'})
    >>> value # doctest: +SKIP
    {'town': u'Arvada', 'user_id': u'1', 'created': u'2008-12-28 17:33:11.078931',
      'user_name': u'asdf', 'town_id': u'1', 'groups': u'4', '_password': '******',
      'password': '******', 'email_address': u'asdf@asdf.com', 'display_name': u'None'}

    """
    def get_value(self, values=None, **kw):
        obj = self.__provider__.get_obj(self.__entity__, params=values, fields=self.__fields__)
        values = self.__provider__.dictify(obj, self.__fields__, self.__omit_fields__)
        for key in self.__fields__:
            method = getattr(self, key, None)
            if method:
                if inspect.ismethod(method):
                    values[key] = method(obj, **kw)
        return values

class RecordFiller(EditFormFiller):pass


class AddFormFiller(FormFiller):
    def get_value(self, values=None, **kw):
        """xxx: get the server/entity defaults."""
        kw = super(AddFormFiller, self).get_value(values, **kw)
        return self.__provider__.get_default_values(self.__entity__, params=values)



"""
saprovider Module

this contains the class which allows dbsprockets to interface with sqlalchemy.

Classes:
Name                               Description
SAProvider                         sqlalchemy metadata/crud provider

Exceptions:
None

Functions:
None

This module is deprecated

Copyright (c) 2007-10 Christopher Perkins
Original Version by Christopher Perkins 2007
Released under MIT license.
"""

from .sa.provider import SAORMProvider as _SAORMProvider, SAORMProviderError as _SAORMProviderError
import warnings

class SAORMProviderError(_SAORMProviderError):
    def __init__(self, *args, **kw):
        warnings.warn('This class has moved to the sprox.sa.provider module.')
        _SAORMProviderError.__init__(self, *args, **kw)

class SAORMProvider(_SAORMProvider):
    def __init__(self, *args, **kw):
        warnings.warn('This class has moved to the sprox.sa.provider module.')
        _SAORMProvider.__init__(self, *args, **kw)




"""Dojo Specific Widgets"""

from tw.dojo import DojoQueryReadStore, DojoBase, grid_css, tundragrid_css, DojoJsonRestStore, buildService
from tw.dojo.selectshuttle import DojoSelectShuttleField, DojoSortedSelectShuttleField
from tw.api import JSSource
from sprox.widgets import PropertyMixin

from tw.dojo import DojoBase, tundragrid_css, tundra_css, dojo_css, dojo_js
from tw.core import JSLink

sprox_grid_js = JSLink(modname="sprox",
                       filename="widgets/tw1widgets/static/dojo_grid.js",
                       )

class SproxDojoGrid(DojoBase):
    engine_name=None
    available_engines = ['mako', 'genshi']
    css = [grid_css, tundragrid_css, tundra_css]
    javascript=[dojo_js, sprox_grid_js]
    dojoType = 'dojox.grid.DataGrid'
    require = ['dojox.grid.DataGrid', 'twdojo.data.TWDojoRestStore']
    params = ['id', 'attrs', 'columns', 'jsId', 'action',
              'rowsPerPage', 'model', 'delayScroll', 'cssclass', 'actions',
              'columnResizing', 'columnReordering', 'column_widths', 
              'default_column_width', 'headers','column_options', 
              'default_column_options','dojoStoreType','dojoStoreWidget',
              'autoHeight'
              ]
    autoHeight="false"
    delayScroll = "true"
    cssclass="sprox-dojo-grid"
    rowsPerPage = 20
    columns = []
    columnReordering = "false"
    columnResizing="false"
    column_widths = {}
    column_options = {}
    default_column_options = {}
    headers = {}
    default_column_width = "30px"
    include_dynamic_js_calls = True
    action='.json'
    model = None
    actions = True
    dojoStoreType = 'dojox.data.QueryReadStore'
    dojoStoreWidget = None
    template = "sprox.widgets.tw1widgets.templates.dojogrid"
    #attrs = {'style':"height:200px"}
    def update_params(self,d):
        d['dojoStoreWidget']=DojoQueryReadStore()
        super(SproxDojoGrid, self).update_params(d)

class DojoJsonRestStoreInstance(JSSource):
    engine_name='mako'
    require = ['twdojo.data.TWDojoRestStore']
    target='.json'
    url = '.json'
    idAttribute = 'id'
    autoSave = 'true'
    source_vars = ["varId","target","url","idAttribute","autoSave"]
    src = """
    dojo.require("twdojo.data.TWDojoRestStore")
    var ${varId}=new twdojo.data.TWDojoRestStore({target:"${target}",autoSave:"${autoSave and 'true' or 'false'}", service:buildService("${url}"),idAttribute:"${idAttribute}"})
                """

class SproxEditableDojoGrid(DojoBase):
    engine_name=None
    available_engines = ['genshi', 'mako']
    css = [grid_css, tundragrid_css]
    require = ['dojox.grid.DataGrid', 'twdojo.data.TWDojoRestStore']
    dojoType = 'dojox.grid.DataGrid'
    params = ['id', 'attrs', 'columns', 'jsId', 'action',
              'rowsPerPage', 'model', 'delayScroll', 'cssclass', 'actions',
              'columnResizing', 'columnReordering', 'column_widths', 'default_column_width', 'headers','column_options', 'default_column_options','dojoStoreType','dojoStoreWidget',
              'autoHeight'
              ]
    delayScroll = "true"
    cssclass="sprox-dojo-grid"
    rowsPerPage = 20
    columns = []
    columnReordering = "false"
    columnResizing="false"
    column_widths = {}
    column_options = {}
    default_column_options = {'editable':"true"}
    headers = {}
    default_column_width = "10em"
    include_dynamic_js_calls = True
    action='.json'
    model = None
    actions = True
    dojoStoreType = 'twdojo.data.TWDojoRestStore'
    dojoStoreWidget = None
    template = "sprox.widgets.tw1widgets.templates.dojogrid"

    def __init__(self,**kw):
        super(SproxEditableDojoGrid, self).__init__(**kw)
        storeName = kw.get('jsId','') or ''
        storeName = storeName+'_store'

        self.javascript.append(buildService)
        self.javascript.append(DojoJsonRestStoreInstance(varId=storeName,target=kw['action'],url=kw['action']))


class SproxDojoSelectShuttleField(DojoSelectShuttleField, PropertyMixin):
    def update_params(self, d):
        self._my_update_params(d)
        super(SproxDojoSelectShuttleField, self).update_params(d)

class SproxDojoSortedSelectShuttleField(DojoSortedSelectShuttleField, PropertyMixin):
    def update_params(self, d):
        self._my_update_params(d)
        super(SproxDojoSortedSelectShuttleField, self).update_params(d)




"""
widgetselecter Module

this contains the class which allows the ViewConfig to select the appropriate widget for the given field

Classes:
Name                               Description
MingWidgetSelecter                Aid for the selection of a widget to correspond to a Ming Document field

Exceptions:
None

Functions:
None


Copyright (c) 2009 Christopher Perkins
Original Version by Christopher Perkins 2009
Released under MIT license.

Mongo Contributions by Jorge Vargas
"""

from sprox.widgetselector import WidgetSelector
from sprox.widgets import *

from ming import schema as S
try:
    from ming.odm.property import RelationProperty, ManyToOneJoin, OneToManyJoin
except ImportError: #pragma: no cover
    from ming.orm.property import RelationProperty, ManyToOneJoin, OneToManyJoin

try:
    from ming.odm.property import ManyToManyListJoin
except: #pragma: no cover
    class ManyToManyListJoin:
        pass

class MingWidgetSelector(WidgetSelector):

    default_multiple_select_field_widget_type = PropertyMultipleSelectField
    default_single_select_field_widget_type = PropertySingleSelectField
    default_name_based_widgets = {}

    default_widgets = {
    S.Bool: SproxCheckBox,
    S.Int: TextField,
    S.Float: TextField,
    S.String: TextField,
    S.DateTime: SproxCalendarDateTimePicker,
    S.Binary: FileField,
    S.Value: Label,
    S.ObjectId: TextField
    }
    def select(self,field):
        if hasattr(field, 'name') and field.name:
            if field.name in self.default_name_based_widgets:
                return self.default_name_based_widgets[field.name]

            if field.name.lower() == 'password':
                return PasswordField

        if isinstance(field, RelationProperty):
            join = field.join
            if isinstance(join, ManyToOneJoin):
                return self.default_single_select_field_widget_type
            if isinstance(join, (OneToManyJoin, ManyToManyListJoin)):
                return self.default_multiple_select_field_widget_type
            raise NotImplementedError("Unknown join type %r" % join)	# pragma: no cover
        
        f = getattr(field, 'field', None)
        if f is not None:
            schemaitem = S.SchemaItem.make(field.field.type)
            if isinstance(schemaitem, S.OneOf):
                return self.default_single_select_field_widget_type
        else:
            return TextField 

        #i don't think this works in the latest ming
        sprox_meta = getattr(field, "sprox_meta", {})
        if sprox_meta.get("narrative"):
            return TextArea
        sprox_meta = getattr(field, 'sprox_meta', None)
        if sprox_meta and 'password' in sprox_meta:
            return PasswordField
        
        return self.default_widgets.get(schemaitem.__class__, TextField)




import platform, sys

if platform.system() == 'Windows': # pragma: no cover
    WIN = True
else: # pragma: no cover
    WIN = False

# True if we are running on Python 3.
PY2 = sys.version_info[0] == 2

if not PY2: # pragma: no cover
    string_type = str
    unicode_text = str
    byte_string = bytes
else:
    string_type = basestring
    unicode_text = unicode
    byte_string = str


try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest